/***************************************************************************
                          CSI.cpp  -  find the CSI index
                             -------------------
    begin                : Mon Aug 21 2000
    copyright            : (C) 2000 by Yuanpeng Janet Huang
    email                : yphuang@cabm.rutgers.edu
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <stdlib.h>

#include "atom.h"
#include "utility.h"

string itoa(int);
typedef map<int, int, less<int> > CSIMap;

#define fabs(x) fabs((double)x)
extern bool DEBUG;
extern void error(char*, char*);
extern void error(string, string);

//-------------------------------------------------
// prepare the inputs:
//     - read the liberay
//     - load the chemical shift
//-------------------------------------------------

// load the csi.ha, csi.co ... libery
typedef map<char, float, less<char> > charMap;
typedef map<int, float, less<int> > intMap;

//global variables -- will be used for gui
//read from libary
charMap csi_ha, csi_cb, csi_ca, csi_c;
charMap r_ha, r_cb, r_ca, r_c;

// for corrections
charMap iso_dummy, iso_ca, iso_cb;


enum atomType { HA, CA, CB, C, CON };
typedef map<atomType, int, less<atomType> > ssMap;

ssMap helix, sheet;

void readCSI(ifstream& in, charMap& aS, charMap& aR, bool isISO, charMap& iso)
{
  string firstLine;
  getline(in, firstLine);

  char AA;
  float shift;
  float range;
  while (in >> AA >> shift >> range)
    {
      if (isISO)
	shift = shift + iso[AA];
      aS.insert(pair<const char, float>(AA, shift));
      aR.insert(pair<const char, float>(AA, range));
    }
}

void readISO(ifstream& in, charMap& aS, charMap& aR)
{
  string firstLine;
  getline(in, firstLine);
  getline(in, firstLine);
  getline(in, firstLine);
  getline(in, firstLine);
  getline(in, firstLine);
  getline(in, firstLine);
  getline(in, firstLine);

  string AAs;
  char AA;
  float HAshift;
  float CAshift;
  float CBshift;
  float Cshift;
  float Nshift;
  float HNshift;
  while (in >> AAs >> AA >> HAshift >> CAshift >> CBshift >> Cshift >> Nshift >> HNshift)
    {
      aS.insert(pair<const char, float>(AA, CAshift));
      aR.insert(pair<const char, float>(AA, CBshift));
    }
}

void initLoad(const Atom_List& atoms, bool isISO) {
  helix[HA] = -1; // define the represention of -1 and 1 for each HA, CA, CB
  sheet[HA] = 1;
  helix[CA] = 1;
  sheet[CA] = -1;
  helix[CB] = -1; // -2;
  sheet[CB] = 1;
  helix[C] = 1;
  sheet[C] = -1;
  helix[CON] = -1;
  sheet[CON] = 1;

	if (!getenv("ASLIB"))
		error("Define environment variable: ASLIB", "");

	string NALIBPath = getenv("ASLIB");

  ifstream isoFile((NALIBPath + "/isoadj.tab").c_str());
  if (!isoFile) error("cannot open liberay file", NALIBPath + "/isoadj.tab");
  readISO(isoFile, iso_ca, iso_cb);
  isoFile.close();

  ifstream haFile((NALIBPath + "/csi.ha").c_str());
  ifstream caFile((NALIBPath + "/csi.ca").c_str());
  ifstream cbFile((NALIBPath + "/csi.cb").c_str());
  ifstream cFile((NALIBPath + "/csi.co").c_str());

  if (!haFile) error("cannot open liberay file", NALIBPath + "/csi.ha");
  readCSI(haFile, csi_ha, r_ha, isISO, iso_dummy);
  if (!caFile) error("cannot open liberay file", NALIBPath + "/csi.ca");
  readCSI(caFile, csi_ca, r_ca, isISO, iso_ca);
  if (!cbFile) error("cannot open liberay file", NALIBPath + "/csi.cb");
  readCSI(cbFile, csi_cb, r_cb, isISO, iso_cb);
  if (!cFile) error("cannot open liberay file", NALIBPath + "/csi.co");
  readCSI(cFile, csi_c, r_c, isISO, iso_dummy);

  haFile.close();
  caFile.close();
  cbFile.close();
  cFile.close();

}

// read the chemical shift list and substract the indexHa, ... information
void readProtein(bool noe, const Atom_List& atoms, intMap& indexHa, intMap& indexCa, intMap& indexCb, intMap& indexC)
{
  float sum = 0; int index = 0;
  for(Atom_List::const_iterator iter = atoms.begin(); iter != atoms.end(); ++iter)
  {
    if (DEBUG) cout << "check CSI " << *iter << endl << iter->get_difforig() << " " << iter->get_shift() << " " << iter->get_i_shift() << endl;

    if (noe && (iter->get_difforig() == -1000000)) // don't depend on the intra assignments
      {
    	if (DEBUG) cout << "noe " << noe << " skip CSI" << endl;
    	continue;
      }

    //if (DEBUG) cout << "check CSI " << *iter << endl << iter->get_difforig() << " " << iter->get_shift() << " " << iter->get_i_shift() << endl;

    if ((*iter).get_name() == "CA")
    	indexCa.insert(pair<const int, float>((*iter).get_index(), (*iter).get_shift()));
    else if ((*iter).get_name() == "CB")
    	indexCb.insert(pair<const int, float>((*iter).get_index(), (*iter).get_shift()));
    else if ((*iter).get_name() == "HA")
    	indexHa.insert(pair<const int, float>((*iter).get_index(), (*iter).get_shift()));
    else if ((*iter).get_name() == "C")
    	indexC.insert(pair<const int, float>((*iter).get_index(), (*iter).get_shift()));
    else if ((*iter).get_name() == "QA")
    	indexHa.insert(pair<const int, float>((*iter).get_index(), (*iter).get_shift()));
    else if (((*iter).get_name() == "HA2") || ((*iter).get_name() == "HA3"))
    {

//    cerr << (*iter) << endl;
      if (sum == 0)
      {
        sum += (*iter).get_shift();
        index = (*iter).get_index();
      }
      else
      {
        if ((*iter).get_index() == index)
	      {
	        sum += (*iter).get_shift();
    		  indexHa.insert(pair<const int, float>((*iter).get_index(), sum/2.0));
    		  sum = 0; index = 0;
	      }
        //else
	//        error("*** CSI: Only one HA found for ", "G" + itoa(index));
      }
    }
  }
}

//--------------------------------------------------------
// CSI filter
//--------------------------------------------------------

// a state machine, convert the raw index to CSI index
enum STATES { START, S1, SN1, PS1, PSN1 };
typedef set<pair<int, int>, less<pair<int, int> > > SESet; //start and end set

void filter(ostream& outputSec, const SEQM& seq, atomType at, CSIMap& raw2, CSIMap& filtered)
{
	SESet sheetSE, helixSE;
  sheetSE.clear();
  helixSE.clear();

  pair<STATES, int> currStat;
  currStat.first = START;
  currStat.second = (*seq.begin()).first;

  CSIMap raw; raw = raw2;
  for(map<int, AA, less<int> >::const_iterator i = seq.begin(); i != seq.end(); i++)
  {
	  if (raw.find((*i).first) == raw.end())
	    {

		  raw.insert(make_pair(i->first, 0)); // chang NA to 0
	    }
  }

  CSIMap::iterator curr;
  for(map<int, AA, less<int> >::const_iterator i = seq.begin(); i != seq.end(); i++)
  {
    curr = raw.find((*i).first);
    if (curr != raw.end())
    {
	    switch (currStat.first)
	    {
        case START:
	        if ((*curr).second == 1)
		        currStat = pair<STATES, int>(S1, (*curr).first);
	        else if ((*curr).second == -1)
		        currStat = pair<STATES, int>(SN1, (*curr).first);
	        else
		        currStat.first = START;
	        break;
        case S1:
	        if ((*curr).second == 0)
		      {
		        if (((*i).second.get_sym_1() == 'P') && (helix[at] == 1)) // proline is not allowed in helix
		        {
		          int length = (*curr).first - currStat.second;
		          if (((sheet[at] == 1) && (length >= 2))|| ((helix[at] == 1) && (length >=4)))
			        {
			          if (sheet[at] == 1)
			            sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
			          else
			            helixSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
			          }
		            currStat = pair<STATES, int>(START, (*curr).first);
		          }
		          else
		            currStat.first = PS1;
		      }
	        else if ((*curr).second == -1)
		      {
		        int length = (*curr).first - currStat.second;
		        if (((sheet[at] == 1) && (length >= 2))|| ((helix[at] == 1) && (length >=4)))
		        {
		          if (sheet[at] == 1)
			          sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
		          else
			          helixSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
		        }
		        currStat = pair<STATES, int>(SN1, (*curr).first);
		      }
	        break;
        case SN1:
	        if ((*curr).second == 0)
		      {
		        if (((*i).second.get_sym_1() == 'P') && (helix[at] == -1)) // proline is not allowed in helix 
		        {
		          int length = (*curr).first - currStat.second;
		          if (((sheet[at] == -1) && (length >= 2))|| ((helix[at] == -1) && (length >=4)))
			        {
			          if (sheet[at] == -1)
			            sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
			          else
			            helixSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
			        }
		          currStat = pair<STATES, int>(START, (*curr).first);
		        }
		        else
		          currStat.first = PSN1;
		      }
	        else if ((*curr).second == 1)
		      {
		        int length = (*curr).first - currStat.second;
		        if (((sheet[at] == -1) && (length >= 2))|| ((helix[at] == -1) && (length >=4)))
		        {
		          if (sheet[at] == -1)
			          sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
		          else
			          helixSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
		        }
		        currStat = pair<STATES, int>(S1, (*curr).first);
		      }
	        break;
	      case PS1:
	        if ((*curr).second == 1)
		        currStat.first = S1;
	        else
	        {
		        int length = (*curr).first - currStat.second -1;
		        if (((sheet[at] == 1) && (length >= 2))||((helix[at] == 1) && (length >=4)))
		        {
		          if (sheet[at] == 1)
		            sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-2));
		          else
		            helixSE.insert(pair<int, int>(currStat.second, (*curr).first-2));
		        }
		        if ((*curr).second == 0)
		          currStat.first = START;
		        else if ((*curr).second == -1)
		          currStat = pair<STATES, int>(SN1, (*curr).first);
	        }
	        break;
	      case PSN1:
	        if ((*curr).second == -1)
		        currStat.first = SN1;
	        else
	        {
		        int length = (*curr).first - currStat.second -1;
		        if (((sheet[at] == -1) && (length >= 2))|| ((helix[at] == -1) && (length >=4)))
		        {
		          if (sheet[at] == -1)
		            sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-2));
		          else
		            helixSE.insert(pair<int, int>(currStat.second, (*curr).first-2));
		        }
		        if ((*curr).second == 0)
		          currStat.first = START;
		        else if ((*curr).second == 1)
		          currStat = pair<STATES, int>(S1, (*curr).first);
	        }
	        break;
        default:
	        break;
      }
    }
   else
	  continue;
  }

  // summary
  //cerr << curr->first;
  if (curr != raw.end())
  {

	    if (DEBUG)
		    outputSec << "currStat.first " << currStat.first << endl;

    switch (currStat.first)
    {
      case S1:
        if (sheet[at] == 1)
  	      sheetSE.insert(pair<int, int>(currStat.second, (*curr).first));
        else
  	      helixSE.insert(pair<int, int>(currStat.second, (*curr).first));
        break;
      case PS1:
        if (sheet[at] == 1)
  	      sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
        else
  	      helixSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
        break;
      case SN1:
        if (sheet[at] == -1)
  	      sheetSE.insert(pair<int, int>(currStat.second, (*curr).first));
        else
  	      helixSE.insert(pair<int, int>(currStat.second, (*curr).first));
        break;
      case PSN1:
        if (sheet[at] == -1)
  	      sheetSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
        else
  	      helixSE.insert(pair<int, int>(currStat.second, (*curr).first-1));
        break;
      default:
        break;
    }
  }
  outputSec << "helix: ";
  for(SESet::iterator i = helixSE.begin(); i != helixSE.end(); i++)
  {
    int sum = 0; int count = 0;
    for(int j = (*i).first; j <= (*i).second; j++)
	  {
	    //cerr << i->first << " " << i->second;
	    if (raw.find(j) != raw.end())
	    {
	      count++;
	      sum += raw[j];
	    }
	  }
	  // if 70% of the segment are helix and at least four
    if ((fabs(count/(float)sum) > 0.7) && (fabs(sum) >4))
	  {
	    for(int j = (*i).first; j <= (*i).second; j++)
	      filtered.insert(pair<int, int>(j, 1));
	    outputSec << (*i).first << "-" << (*i).second << " ";
	  }
  }
  outputSec << endl;

  outputSec << "sheet: ";
  for(SESet::iterator i = sheetSE.begin(); i != sheetSE.end(); i++)
  {

	  if (DEBUG)	outputSec << "sheetSE" <<  i->first << " " << i->second << endl;

    int sum = 0; int count = 0; int conti = 0; int max = 0;
    for(int j = (*i).first; j <= (*i).second; j++)
    {
      if (raw.find(j) != raw.end())
      {
        if (raw[j])
   	      conti++; // continuously sheet
        else
   	    {
   	      if (conti > max) max = conti;
   	        conti = 0;
   	    }
         count++;
         sum += raw[j];
      }
    }
    if (conti > max) max = conti;

	  // if 70% of the segment are sheet and continuously have 3

    // now, change it to 2
    if ((fabs(count/(float)sum) > 0.7) && ((max >=2) || (fabs(sum) >= 2)))
    {
      for(int j = (*i).first; j <= (*i).second; j++)
        filtered.insert(pair<int, int>(j, -1));
      outputSec << (*i).first << "-" << (*i).second << " ";
    }
  }
  outputSec << endl;
}

// combine ca and cb CSI
void consensus(const SEQM& seq, CSIMap& indexCaFilter, CSIMap& indexCaRaw, CSIMap& indexCbFilter, CSIMap& indexCbRaw, CSIMap& indexFilter, ostream& outputSec)
{
  for(SEQM::const_iterator i = seq.begin(); i != seq.end(); i++)
  {
    if (indexCaFilter[(*i).first] == 1) // for helix
      {
	if ((indexCaRaw.find(i->first) != indexCaRaw.end()) && (indexCbRaw.find(i->first) != indexCbRaw.end()))
	  {
	    if ((indexCaRaw[i->first] != 1) && (indexCbRaw[i->first] != -1))
	      indexFilter[i->first] = 0; // exclude appreatly loops
	    else
	      indexFilter[(*i).first] = 1;
	  } else
	    indexFilter[i->first] = 1;
      }
    else if ((indexCaFilter[(*i).first] == -1)&&((indexCbRaw[i->first] == 1)||(indexCbFilter[(*i).first]==-1))) // only use Ca for beta
	indexFilter[(*i).first] = -1;
    //else if ((indexCbFilter[(*i).first] == -1) && (indexCaFilter[(*i).first] == -1)) // beta, use both ca and cb
    //  indexFilter[(*i).first] = -1;
    else if ((indexCaRaw.find(i->first) == indexCaRaw.end()) &&  // not used anymore - if ca missing - fill in with ca/ha see checkCb
    	     (indexCbFilter.find(i->first) != indexCbFilter.end()))
      indexFilter[i->first] = indexCbFilter[i->first];
    else
      indexFilter[(*i).first] = 0;
  }

  // remove buldge
  for(SEQM::const_iterator i = seq.begin(); i != seq.end(); i++)
	{
		if (indexFilter[i->first] == 0)
		{
			if ((indexFilter.find(i->first-1) != indexFilter.end()) && (indexFilter.find(i->first+1) != indexFilter.end()))
			{
				if ((indexFilter[i->first-1] == -1) && (indexFilter[i->first+1] == -1) && (indexCaRaw[i->first]==-1))
				{
					indexFilter[i->first] = -1;
					//if (DEBUG)
					//outputSec << indexCaFilter[i->first] << " " << i->first << "set to -1" << endl;
				}
			}
		}
	}

  // helix - must have 11 on ends
  for(SEQM::const_iterator i = seq.begin(); i != seq.end(); i++)
	{
	  if ((indexFilter[i->first] == 1)) // helix  case
	    {
	      if ((indexFilter.find(i->first-1) != indexFilter.end()) && (indexFilter[i->first-1] == 1) &&
		  (indexFilter.find(i->first-2) != indexFilter.end()) && (indexFilter[i->first-2] != 1))
		{
		  if ((indexCaRaw[i->first] != 1)) // 10
		    {
		      indexFilter[i->first] = 0;
		      indexFilter[i->first-1] = 0;
		      //if (DEBUG)
		      //outputSec << indexCaFilter[i->first] << " " << i->first << "set to -1" << endl;
		    }
		}
	      if ((indexFilter.find(i->first+1) != indexFilter.end()) && (indexFilter[i->first+1] == 1) &&
		  (indexFilter.find(i->first+2) != indexFilter.end()) && (indexFilter[i->first+2] != 1))
		{
		  if ((indexCaRaw[i->first] != 1)) // 01
		    {
		      indexFilter[i->first] = 0;
		      indexFilter[i->first+1] = 0;
		      //if (DEBUG)
		      //outputSec << indexCaFilter[i->first] << " " << i->first << "set to -1" << endl;
		    }
		}
	    }
	}
}

// find raw index
void rawIndex(const SEQM& seq, intMap& indexA, charMap& csiA, charMap& rA, CSIMap& rawA)
{
  float pv = 0; // privious one
  float pid = 0;
  for(intMap::iterator i = indexA.begin(); i != indexA.end(); i++)
  {
    if (DEBUG) cout << i->first << " " << i->second << endl;

    float t = (*i).second - csiA[seq.find((*i).first)->second.get_sym_1()];


   //fix the cys
    if (seq.find(i->first)->second.get_sym_1() == 'C')
    {
      float tb = (*i).second - csiA['B'];

      if (fabs(tb) < fabs(t))
	t = tb;
    }
    //---end

    if (t > rA[seq.find((*i).first)->second.get_sym_1()] + 0.00001)
	    rawA.insert(make_pair((*i).first, 1));
    else if ( t < -1 * rA[seq.find((*i).first)->second.get_sym_1()] - 0.00001)
	    rawA.insert(make_pair((*i).first, -1));
    // t is at the edge and between
    else if ((((t > 0) && (pv < 0)) || ((t < 0) && (pv > 0))) && (i->first - pid == 1)) // if changeed sign
	  {
	    if (t > rA[seq.find((*i).first)->second.get_sym_1()] - 0.00001) // at the edge
	      rawA.insert(make_pair((*i).first, 1));
	    else if ( t < -1 * rA[seq.find((*i).first)->second.get_sym_1()] + 0.00001) // at the edge
	      rawA.insert(make_pair((*i).first, -1));
	    else
	      rawA.insert(make_pair((*i).first, 0));
	  }
    else
	    rawA.insert(make_pair((*i).first, 0));
    pv = t;
    pid = i->first;
  }
}

void checkCb(const SEQM& seq, CSIMap& indexHaRaw, CSIMap& indexCaRaw, CSIMap& indexCbRaw)
{
  for(map<int, AA, less<int> >::const_iterator i = seq.begin(); i != seq.end(); i++)
    {

      if (indexCbRaw.find(i->first) == indexCbRaw.end()) // not available
	{
		if ((indexCaRaw[i->first] == 0) || (indexHaRaw[i->first] == 0))
		{
		  	indexCbRaw[i->first] = (indexHaRaw[i->first] - indexCaRaw[i->first]);
		} else // use Ha values or both
		{
		  indexCbRaw[i->first] = (indexHaRaw[i->first] - indexCaRaw[i->first])/2;
		}
	}

      if (indexCaRaw.find(i->first) == indexCaRaw.end()) // not available
	{
		if ((indexCbRaw[i->first] == 0) || (indexHaRaw[i->first] == 0))
		{
		  	indexCaRaw[i->first] = -1*(indexHaRaw[i->first] + indexCbRaw[i->first]);
		} else
		{
		  indexCaRaw[i->first] = -1 * (indexHaRaw[i->first] + indexCbRaw[i->first])/2;
		}

	}
    }
}


//////////////////////
// main function    //
//////////////////////
void CSI_method(bool noe, ostream& outputSec, const SEQM& seq, const Atom_List& atoms, CSIMap& indexFilter, CSIMap& indexCaRaw, CSIMap& indexCbRaw, bool isCSI, bool isISO)
{

  intMap indexHa, indexCa, indexCb, indexC;

  if (!isCSI)
    {
      for(SEQM::const_iterator ii = seq.begin(); ii != seq.end(); ii++)
	{
	  indexFilter[ii->first] = 0;
	}
      return;
    }
  outputSec << "CSI Method" << endl;
  outputSec << "==========" << endl << endl;

  initLoad(atoms, isISO);
  readProtein(noe, atoms, indexHa, indexCa, indexCb, indexC);

	// the raw -1,1,0 information, for each HA, CA, CB ..,
	CSIMap indexHaRaw, indexCRaw, rawCon;
	// the -1,1,0 information (CSI)
	CSIMap indexHaFilter, indexCaFilter, indexCbFilter, indexConFilter, indexCFilter;

  // create raw index
  rawIndex(seq, indexHa, csi_ha, r_ha, indexHaRaw);
  if (DEBUG) cout << "Ca-Index " << endl;
  rawIndex(seq, indexCa, csi_ca, r_ca, indexCaRaw);
  rawIndex(seq, indexCb, csi_cb, r_cb, indexCbRaw);

  checkCb(seq, indexHaRaw, indexCaRaw, indexCbRaw);
  //rawConsensusIndex();

  // create CSI index
  outputSec << "Ha-Index : " << endl;
  filter(outputSec, seq, HA, indexHaRaw, indexHaFilter);
  outputSec << "Ca-Index : " << endl;
  filter(outputSec, seq, CA, indexCaRaw, indexCaFilter);
  outputSec << "Cb-Index : " << endl;
  filter(outputSec, seq, CB, indexCbRaw, indexCbFilter);
  outputSec << "Summation : " << endl;
  filter(outputSec, seq, CON, rawCon, indexConFilter);

  consensus(seq, indexCaFilter, indexCaRaw, indexCbFilter, indexCbRaw, indexFilter, outputSec);

  outputSec << endl << "Summary:" << endl;
  outputSec << "#\tAA\tHA\tCA\tCB\tSUM\tCONSENSUS(CA/CB)"<< endl;
   outputSec << "-\t--\t--\t--\t--\t---\t----------------" << endl;
  for(SEQM::const_iterator ii = seq.begin(); ii != seq.end(); ii++)
  {
    outputSec << (*ii).first << "\t" << (*ii).second.get_sym_1() << "\t";

    CSIMap::iterator i;
    i = indexHaRaw.find((*ii).first);
    if (i != indexHaRaw.end())
	    outputSec << (*i).second << "\t";
    else
	    outputSec << "NA\t";

    i = indexCaRaw.find((*ii).first);
    if (i != indexCaRaw.end())
	    outputSec << (*i).second << "\t";
    else
	    outputSec << "NA\t";

    i = indexCbRaw.find((*ii).first);
    if (i != indexCbRaw.end())
	    outputSec << (*i).second << "\t";
    else
	    outputSec << "NA\t";

    i = rawCon.find((*ii).first);
    if (i != rawCon.end())
	    outputSec << (*i).second << "\t";
    else
	    outputSec << "NA\t";

    outputSec << indexFilter[(*ii).first];
    outputSec << endl;
  }
}
