/***************************************************************************
                          atom.h  -  
                             -------------------
    begin                : Mon Aug 21 2000
    copyright            : (C) 2000 by Yuanpeng Janet Huang
    email                : yphuang@cabm.rutgers.edu
 ***************************************************************************/

#ifndef ATOMS_H
#define ATOMS_H

#include <string>
#include <set>
#include <list>
#include <deque>
#include "AA.h"
#include "SmplStat.h"

using namespace std;

/*********************************
pseudo: for pseudo atom correction
*********************************/
struct pseudo {
	string  pname; // pseudo name 
	float 	cerct; // correction 
	string 	cment; // comment
	int 	div;   // divide

	pseudo() { cerct = 0; cment = ""; div = 1; pname = ""; };
	void reset() { cerct=0; cment=""; div=1; pname="";};
};

/***********************
atom: object of the atom 
one atom corresponds to one
resonance assignment in the bmrb file
***********************/

class atom {
protected:
  float errorA; // the usage is not implmented
  float  shift;     // chmical shift value
	float  HCCHshift; // chmical shift value in HCCHCOSY, local refined value
  string name;      // name of the atom, stantard name, given by the user
  char   type;      // H, C, O , S ...
	int    ambi; 	  // ambiguity
  int    index;     // index of the AA ;
	int    atom_ID;   // index of the atom; 
	char   symbol_1;  // symbol of AA

  atom* attached_N; // pointer to the attached_N atom,
                          // that declared elseware. no new or delete here    
  deque<atom*> attached_C; // queue of pointers to the attached_C atom

  mutable string layer;
	atom* _repAtom;
	
public:
	//--- init
  atom();
	~atom(){}; 
  atom(int);                                  // index
	atom(int, int, string, char, float, int);   // atom_ID, index, user-name, type, shift, ambi
	void set(int, int, string, char, float, int);    // atom_ID, index,user-name, type, shift, ambi

	//--- io 
	friend ostream& operator<<(ostream&, const atom&);
	friend ostream& operator<<(ostream&, const atom*);
	friend istream& operator>>(istream&, atom&);
	friend bool operator<(const atom&, const atom&);
  friend bool operator!=(const atom&, const atom&);
  friend bool operator==(const atom&, const atom&);
	
  //--- check if attached to N or C 	
  bool isatt_N() const;
	bool isatt_C() const;
	atom* const att_N() const;
	const deque<atom*>& att_C() const;
	atom* const att_C(float, float) const; // given a chemical shift, find the closest matched one
	atom* const att_X(float, float) const;

	void set_repAtom(atom*);
  atom* get_repAtom() const;
	int ambiDegree() const;
	
  //--- check atom type
	bool isH() const;
	bool isN() const;
	bool isC() const;
	bool isUnique()const; // is uniquely assignmed, no any ambiguious (including methyl group)
	bool isDe(const atom*) const; // this atom and atom* are degenerate or not?
	bool isDeC(const atom*) const;
	bool less4(const atom*) const;	
	bool length3(const atom*) const;
	bool isRingAtom() const;
	bool isMethyleneAtom() const;
	bool isMeaningless(const atom*) const;
	bool isSecLongHNHN(const atom*) const; 
 	bool MethylFiveBonds(const atom*) const;
 //--- return the residue it in.
	const AA& residue() const;
	int    get_ID() const;

  float  get_shift() const;

	float  get_Xdiff(float, float) const;
	float  get_Hdiff(float, float) const;
 	float  get_XdiffAb(float, float) const;
	float  get_HdiffAb(float, float) const;
	float  get_difforig() const;
	
  float  get_i_shift() const;
  string get_name() const;
  int    get_index() const;
  string get_layer() const;
  void set_layer(string);
  void set_symbol_1(char);
  
  char   get_type() const;
	int    get_ambi() const;
	int    get_deg() const;
	int 	get_degCount() const;
	
	string get_resi_1() const;
	char   get_sym_1() const;

	void   set_att_N(atom*);
	void   set_att_C(atom*);
	void   set_name(const string&);
	void   set_atom_ID(int);
	void   set_HCCHshift(float);
  void   set_ambi(int);
  void   set_shift(float);

  void   resetLayer() const { layer = "NA"; };
	//--- for pseudo
	void get_pseudoboth(pseudo&) const; // if both peak appearce
	void get_pseudo(const atom*, pseudo&) const; // if only one peak
	void get_pseudo2(const atom*, pseudo&) const; // if only one peak
	void get_pseudoIC(const atom*, pseudo&) const;
	void get_pseudoRaw(const atom*, pseudo&, bool) const;
};

typedef pair<atom* const, atom* const> MomDad;
typedef pair<const atom* const, const atom* const> cMomDad;

typedef set<MomDad, less<MomDad> > MomDadSet;
typedef set<cMomDad, less<cMomDad> > cMomDadSet;

typedef list<atom> Atom_List;
typedef set<const atom*, less<const atom*> > AtomSet;

// a list of atoms by the atom_ID
typedef map<int, atom*, less<int> > Atom_ID_List;

// a list of atoms by its shift value, this is generated with atoms
typedef multimap<float, atom*, less<float> > Atom_Shift_List;
typedef multimap<int, atom*, less<int> > Atom_Index_List; // by the AA index

//
typedef map<const cMomDad, SampleStatistic, less<const cMomDad> > disAtomPairStat;
//typedef map<const cMomDadSet, SampleStatistic, less<const cMomDadSet> > disMomDadSetStat;
typedef map<const cMomDad, pair<cMomDadSet, float>, less<const cMomDad> > disMomDadSetStat;

void make_att_list(ostream&, Atom_List&, float, float);

typedef map<int, char> HisMap; 

#endif

