/***************************************************************************
                          hbond.cpp  -  methods to identify h-bond
                             -------------------
    begin                : Thu Oct 19 2000
    copyright            : (C) 2000 by Yuanpeng Janet Huang
    email                : yphuang@cabm.rutgers.edu
 ***************************************************************************/

#include <cmath>
#include <stdlib.h>
#include "noesy_assign.h"

extern bool DEBUG;

#define fabs(x) fabs((double) x)
#define abs(x) abs((double) x)
#define fmod(x,y) fmod((double) x, (double) y)

ostream& operator<<(ostream& s, const HBOND& hbond)
{
  s << hbond.first << " " << hbond.second << endl;
}


////////////////////////////////////////////
//                                        //
// Read in slow NH exchange rate data     //
//                                        //
////////////////////////////////////////////
istream& operator>> (istream& s, NHSlowList& l)
{
  int i;
  string r;
  while( s >> i >> r)
  {
    if (seq.find(i)->second.get_sym_1() == symbol3_1[r])
      l.insert(i);
    else
      error("In NHSlowList, the " + itoa(i) + "th residue " + r + " is not match to the one in sequence", "");
  }

  ios::iostate state = s.rdstate();

	if ((state&ios::failbit) && !(state&ios::eofbit))
	{
		cerr << endl << "*** Bad NH-exchange format found: " << endl
				 <<  i  << " " << r  << " Stop read" << endl;
	}

  return s;
}

///////////////////////////////////////////
//                                       //
// Find H-bond                           //
//                                       //
///////////////////////////////////////////
void NOESY_Assign::findHbond(int vio, const string fileName, bool dyana, SS& auto_ss, deque<HBOND>& anti_register, deque<HBOND>& para_register,
                const NHSlowList& NHList, matrix& acfMax, set<HBOND, less<HBOND> >& HBONDList)
{
  if (DEBUG) NOESY_Assign::outputLog() << "test in findHbond " << fileName << " " << dyana<< endl;


  ///////////////////////////////////////////////
  //----- iterative stage using 3D structure   //
  ///////////////////////////////////////////////

  multimap<int, pair<int, float>, less<int> > s3dHBOND; s3dHBOND.clear(); // O-H
  multimap<int, pair<int, float>, less<int> > s3dHBOND2; s3dHBOND2.clear(); // H-O

  bool ovw = 0;

  if (fileName != "")
  {
  // read in from ovw files --- find all h-bonds that idendified by DYANA
  ifstream fovw(fileName.c_str());

  if (DEBUG) NOESY_Assign::outputLog() << dyana << "-" << endl;

  if ((fovw) && dyana) // dyana
  {
  	ovw = 1;
    string line;
    while(getlineNA(fovw, line) && (line.find("Hydrogen bonds") > line.length())) // != "Hydrogen bonds"))
    {
      if (DEBUG) NOESY_Assign::outputLog() << "scanning " << line << endl;
    }

     while(getlineNA(fovw, line) && (line.find("hydrogen bonds.") > line.length())) // between "Hydrogen bonds:" and "hydrogen bonds".
    {
      if (DEBUG) NOESY_Assign::outputLog() << "scanning++ " << line << endl;

      vector<string> subline; subline.clear();
      split(line, subline);

      if (subline.size() < 8)
	continue;

      //if (DEBUG) NOESY_Assign::outputLog() << "scanning++ " << line << endl;

      float frq = (atoi(subline[7].c_str())/(float)NOESY_Assign::pdbList().size());
      if (( frq > 0.1) &&
            ((subline[0] == "HN") || (subline[0] == "H")) &&
            (subline[4] == "O"))
      {

	if (DEBUG) NOESY_Assign::outputLog() << "read in " << subline[6] << " " << subline[2] << " " << subline[7] << " " << NOESY_Assign::pdbList().size() << endl;
        {

            s3dHBOND.insert(make_pair(atoi(subline[6].c_str()), make_pair(atoi(subline[2].c_str()), frq)));
            s3dHBOND2.insert(make_pair(atoi(subline[2].c_str()), make_pair(atoi(subline[6].c_str()), frq)));
        }
      }
    }
  }
	else if (fovw && !dyana) // from pdbstat
	{
  	if (DEBUG) NOESY_Assign::outputLog() << "in !dyana " << endl;
		ovw = 1;
    string line;
    while(getlineNA(fovw, line) && (line.find("hydrogen bonds") > line.length())) // between "Hydrogen bonds:" and "hydrogen bonds".
    {
      vector<string> subline; subline.clear();
      split(line, subline);

      if (subline.size() == 0)
      	continue;

      if (subline.size() < 10)
        continue;

     float frq = (atoi(subline[9].c_str())/(float)NOESY_Assign::pdbList().size());

      if (((frq > 0.1) &&
            (subline[0] == "H") &&
            (subline[5] == "O")))
      {
        {
	  if (DEBUG) NOESY_Assign::outputLog() << "read in " << subline[8] << " " << subline[3] << " " << subline[9] << " " << NOESY_Assign::pdbList().size() << endl;
            s3dHBOND.insert(make_pair(atoi(subline[8].c_str()), make_pair(atoi(subline[3].c_str()), frq)));
            s3dHBOND2.insert(make_pair(atoi(subline[3].c_str()), make_pair(atoi(subline[8].c_str()), frq)));
        }
      }
    }
	} else
	  {
	    if (DEBUG) NOESY_Assign::outputLog() << "cann't read " << fileName << endl;
	  }
  }
    ///////////////
    // b-stands  //
    ///////////////

    //-- step1: find all possible HBONDs from secondary structure
    multimap<int, int, less<int> > ssHBOND; ssHBOND.clear();

    // antiparallel
    for(deque<HBOND>::iterator iterH=anti_register.begin(); iterH!=anti_register.end();iterH++)
    {
      const Chain& chain1 = seq.find((*iterH).first)->second.get_ss()->get_chain();
      const Chain& chain2 = seq.find((*iterH).second)->second.get_ss()->get_chain();

      // if both are not in the beta, skip, also coded in  EC.cpp
      if ((seq.find(iterH->first)->second.get_ss()->get_type() != -1) && (seq.find(iterH->first)->second.get_ss()->get_type() != -1))
        continue;

      // find the co-starting point
      Chain::reverse_iterator j = (Chain::reverse_iterator) chain2.rbegin();
      Chain::iterator i = chain1.begin();

      int di = abs((*i)-(*iterH).first);
      int dj = abs((*j)-(*iterH).second);

      int d;
      if (di < dj)
      {
        d = di;
        while (j != chain2.rend())
	      {
	        if (fabs((*j)-(*iterH).second) >= d+1)
	          j++;
	        else
	          break;
	      }
      }
      else
      {
        d = dj;
        while (i != chain1.end())
	      {
	        if (fabs((*i)-(*iterH).first) > d)
	          i++;
	        else
	          break;
	      }
      }

      // identify the ends of b-strands
      int n1 = *(chain1.begin());
      int n2= *(chain2.rbegin());
      int c1 = *(chain1.rbegin());
      int c2 = *(chain2.begin());

      //starting walking
      while (i != chain1.end())
      {
        if (j != chain2.rend())
	{
	        if (fmod(d, 2) == 1)
	        {
	          i++;
	          j++;
	          d = 0;
	        }
	        else
	        {
			//if (((*i) != n1) && ((*j) != n2)) //exclude the first one
			{
		        	ssHBOND.insert(make_pair((*i), (*j)));
			}
			//if (((*i) != c1) && ((*j) != c2)) //exclude the last one
			{
		        	ssHBOND.insert(make_pair((*j), (*i)));
			}
	          i++;
	          j++;
	          d = 1;
	        }
	} else
	    break;
      }
    }

    for(deque<HBOND>::iterator iterH=para_register.begin(); iterH!=para_register.end();iterH++)
    {
      const Chain& chain1 = seq.find((*iterH).first)->second.get_ss()->get_chain();
      const Chain& chain2 = seq.find((*iterH).second)->second.get_ss()->get_chain();

      // if both are not in the beta, skip, also coded in EC.cpp
      if ((seq.find(iterH->first)->second.get_ss()->get_type() != -1) && (seq.find(iterH->first)->second.get_ss()->get_type() != -1))
        continue;

	    // find the co-starting point
	    Chain::const_iterator j = chain2.begin();
	    Chain::const_iterator i = chain1.begin();

	    int di = abs((*i)-(*iterH).first);
	    int dj = abs((*j)-(*iterH).second);
	    int d;
	    if (di < dj)
	    {
	      d = di;
	      while (j != chain2.end())
		    {
		      if (fabs((*j)-(*iterH).second) > d)
		        j++;
		      else
		        break;
		    }
	    }
	    else
	    {
	      d = dj;
	      while (i != chain1.end())
		    {
		      if (fabs((*i)-(*iterH).first) > d)
		        i++;
		      else
		        break;
		    }
	    }

  	    // identify the ends of b-strands
// 	ends of b-strands

	    //starting walking
	    while (i != chain1.end())
	    {
	      if (j != chain2.end())
		    {
		      if (fmod(d, 2) == 1)
		      {
		        j++;
		        if (j != chain2.end())
			{
				ssHBOND.insert(make_pair((*j), (*i)));
			}
		        i++;
		        d = 0;
		      }
          	      else
          	      {
		        i++;
			if (i != chain1.end())
			{
				ssHBOND.insert(make_pair((*i), (*j)));
			}
		        j++;
		        d = 1;
		      }
		} else
		    break;
      	   }
    }

  ///////////////////////////////////////
  // main part                         //
  ///////////////////////////////////////

    multiset<int, less<int> > doners; doners.clear();
    multiset<int, less<int> > acceptors; acceptors.clear();

    ///////////////////
    // b-sheet       //
    // Sec + NH or3D //
    ///////////////////

    //--- filter the ssHBOND by NH
    multimap<int, int, less<int> > ssNHHBOND; ssNHHBOND.clear();
    for(multimap<int, int, less<int> >::iterator i=ssHBOND.begin(); i!=ssHBOND.end(); i++) // H-O
    {
        acceptors.insert(i->second);
        if (NHList.empty() && (!ovw))
        {
          //cerr << endl << "***Warning: sheet Hbond " << i->second << "<-" << i->first << " are derived based on NOEs alone";
        } else if ((NHList.find(i->first) == NHList.end()) &&
		   (NOESY_Assign::ddcycle() > 2) &&
                    ((s3dHBOND2.count(i->first) != 1) || (s3dHBOND.count(i->second) != 1) ||
                    (s3dHBOND.find(i->second)->second.first != i->first)))
        {
          continue;
        }
        ssNHHBOND.insert(make_pair(i->second, i->first)); //O-H
        doners.insert(i->first);
    }

    /////////////////////
    // helix           //
    // Sec + NOE or 3D //
    /////////////////////
    for(SS::iterator i = auto_ss.begin(); i!= auto_ss.end(); i++)
    {
      if (i->second.get_type() != 1)
        continue;

      for (Chain::iterator c = i->second.get_chain().begin(); c != i->second.get_chain().end(); c++)
      {
      	// excluding the first and the last one
      	if (*c == *(i->second.get_chain().begin())+4)
	  continue;

	if (*c == *(i->second.get_chain().rbegin()))
	  continue;

        if ((seq.find((*c)-4)->second.get_ss() == &(i->second)))// && (((*c)-4 > 70) || ((*c)-4 < 62)))
      {
        ssHBOND.insert(make_pair(*c, (*c)-4));
        acceptors.insert((*c)-4);

        // if dan(i,i+4) found, alpha-helix
        bool haveDan4 = 0;
      	for(deque<NOESY*>::iterator iter = NOEPeakList.begin(); iter != NOEPeakList.end(); iter++)
	        if (((*iter)->Type() != "2D") && ((*iter)->Type().find("IC") >= (*iter)->Type().length()))
	        {
            haveDan4 += (*iter)->Find_possHaHnNoe(4, *c);
          }

        if (NOESY_Assign::MINCON())
      	for(deque<NOESY*>::iterator iter = NOEPeakList.begin(); iter != NOEPeakList.end(); iter++)
	        if (((*iter)->Type() != "2D") && ((*iter)->Type().find("IC") >= (*iter)->Type().length()))
	        {
		  haveDan4 += (*iter)->Find_possHnHnNoe(2, *c);
		}

	if (DEBUG) NOESY_Assign::outputLog() << "helix " << *c << " haveDan4: " << haveDan4 << endl;

        {
 	  if (//(NHList.find(*c) == NHList.end()) &&
	      (!haveDan4) &&  // no noe to support and strucure to support
	      (NOESY_Assign::ddcycle() > 2) &&
	      (((s3dHBOND2.count(*c) != 1) || (s3dHBOND.count(*c-4) != 1))
	       || (s3dHBOND.find(*c-4)->second.first != *c)))
	    {
	      if (DEBUG) NOESY_Assign::outputLog() << "ddcycle>2 " << *c << " " << NOESY_Assign::ddcycle() << endl;
	      continue;
	    }

          ssNHHBOND.insert(make_pair((*c)-4, *c));
          doners.insert(*c); // H
        }

      }
      }
    }

    if (DEBUG) NOESY_Assign::outputLog() << "cycle " << NOESY_Assign::ddcycle() << "-" << NOESY_Assign::dcycle() << endl;

    if (vio) // has violations - only remove
    {
      if (DEBUG) NOESY_Assign::outputLog() << "in unassigned cycle " << endl;

      set<HBOND, less<HBOND> > hbond2del; hbond2del.clear();
      for(set<HBOND, less<HBOND> >::iterator i = HBONDList.begin(); i!= HBONDList.end(); i++)
	{
	  if ((ssNHHBOND.count(i->first) > 1) || (doners.count(i->second) > 1))
	    {
	      hbond2del.insert(*i);
	    } else if ((s3dHBOND.count(i->first) > 1) || (s3dHBOND2.count(i->second) > 1))
	    {
	      hbond2del.insert(*i);
	    }
	}
      for(set<HBOND, less<HBOND> >::iterator i=hbond2del.begin(); i!=hbond2del.end(); i++)
	{
	  if (DEBUG) NOESY_Assign::outputLog() << "rm " << i->first << " " << i->second << endl;
	  HBONDList.erase(*i);
	}
      return;
    }

    // no violatons - use the strucutre for hbonds
  HBONDList.clear();

    //HBONDList.insert(HBONDList.begin(), ssNHHBOND.begin(), ssNHHBOND.end());
  for(multimap<int, int, less<int> >::iterator i = ssNHHBOND.begin(); i != ssNHHBOND.end(); i++) // O-H
  {
    if (DEBUG) NOESY_Assign::outputLog() << i->first << " " << ssNHHBOND.count(i->first) << " " << doners.count(i->second) << endl;

    if ((ssNHHBOND.count(i->first) == 1) && (doners.count(i->second) == 1)) // doner and accepor are both unique
    {
      if (DEBUG) NOESY_Assign::outputLog() << "From SS " << i->first << " " << i->second << endl;
      HBONDList.insert(*i); // O-H
    }
  }

  /////////////////////////////////
  // for other type of hybonds   //
  // NH, 3D and NOE              //
  /////////////////////////////////

  // if unique, add to hbond list -- none regular structure related hbonding
  for(multimap<int, pair<int, float>, less<int> >::iterator i = s3dHBOND.begin(); i != s3dHBOND.end(); i++) // O-H
	{
if (DEBUG) NOESY_Assign::outputLog() << "hbond-FRQ: " << i->second.second << " " << i->first << " " << i->second.first << endl;
 if (((i->second.second > 0.9) || (NHList.find(i->second.first) != NHList.end())) &&
     (s3dHBOND.count(i->first) == 1) && (s3dHBOND2.count(i->second.first) == 1) && // O, H
     (abs(i->first - i->second.first) > 2) && // have to be at least 3 residues appart
     ((acfMax.find(make_pair(i->first, i->second.first)) != acfMax.end()) || (acfMax.find(make_pair(i->second.first, i->first)) != acfMax.end()) ||
      (acfMax.find(make_pair(i->first-1, i->second.first)) != acfMax.end()) || (acfMax.find(make_pair(i->first+1, i->second.first)) != acfMax.end()) ||
      (acfMax.find(make_pair(i->second.first-1, i->first)) != acfMax.end()) || (acfMax.find(make_pair(i->second.first+1, i->first)) != acfMax.end()) ||
      (acfMax.find(make_pair(i->first, i->second.first-1)) != acfMax.end()) || (acfMax.find(make_pair(i->first, i->second.first+1)) != acfMax.end()) ||
      (acfMax.find(make_pair(i->second.first, i->first-1)) != acfMax.end()) || (acfMax.find(make_pair(i->second.first, i->first+1)) != acfMax.end())) && // found assigned NOE contact
     (ssHBOND.find(i->second.first) == ssHBOND.end()) && // doner and accepor are not shown in the SS,
     (acceptors.find(i->first) == acceptors.end()) &&
     (doners.find(i->second.first) == doners.end()) &&
     (ssNHHBOND.find(i->first) == ssNHHBOND.end())) // doner and accepor are both unique
    {
      HBONDList.insert(make_pair(i->first, i->second.first));
      if (DEBUG) NOESY_Assign::outputLog() << "From Pdb " << i->first << " " << i->second.first << endl;
    }
  }
}
