/*$Id: RadiusOfGyration.c,v 1.8 2003/10/06 12:13:32 roberto Exp roberto $*/
#include "pdbstat.h"
#include "PdbStatMath.h" 

#define Avogadro (6.023e23)

/* ========================================================================= */
void RadiusOfGyration(int Model, double *Xaxe, double *Yaxe, double *Zaxe)
/*
  ---
  --- Compute: 
  		center of masses -- is recomputed for selected atoms
		radius of gyration
		matrix of inertia moments  (CABM, 2002)
		estimation of alignment matrix (CABM, 2002)
  --- (Roberto Tejero, RTT, 2003, CABM-Rutgers University).
  ---
*/
{
  static int    iat, jat, iatCt = 0;
  static float  x_cm, y_cm, z_cm;
  static float  masstot, xtot, ytot, ztot;
  static float  radius_of_gyration;
  static double rgi, rgitot;
  static double Ixx , Iyy, Izz, Ixy, Ixz, Iyz, Inertia[3][3] ;
  static double Gxx , Gyy, Gzz, Gxy, Gxz, Gyz, Gyration[3][3] ;
  static double A11, A22, A33, Axx, Ayy, Azz, I_Rhomb, G_Rhomb ;
  struct models *model ;
  struct atom_pdb *Atom ;

/* To use LAPACK routines one have to define this */
#ifdef LAPACK
   LWORK = 4;
#endif
   Three1 = 3;
   Three2 = 3;

   /* Initalize variables for safety */
  x_cm = 0.0; y_cm = 0.0; z_cm = 0.0;
  xtot = 0.0; ytot = 0.0; ztot = 0.0;
  masstot = 0.0; rgitot = 0.0;
  iatCt = 0 ;

  model = MODELS + Model ;

  if ( !model->Selected ) {
	  fprintf(stdout,"> Rgyration: *NO* model/atom selected, select something first, please!\n");
	  fprintf(stdout,"> Rgyration: using command select, i.e.  select * * * \n");
	  return;
  }
  
  /*  Check if we have a molecule read anyway */
  if (model->number_of_atoms <= 0) {
    printf("> INFO_mol: ** No molecule read ** \n");
    return;
  }
  
  /* Get center of masses of selected atoms */
  for(iat = 1 ; iat <= model->number_of_atoms ; iat++) {
    Atom = model->PDB + iat ;
    if ( Atom->Selected ) {
      xtot = xtot + Atom->pdb_masa * Atom->pdb_xcrd ;
      ytot = ytot + Atom->pdb_masa * Atom->pdb_ycrd ;
      ztot = ztot + Atom->pdb_masa * Atom->pdb_zcrd ;
      masstot = masstot + Atom->pdb_masa ;
      iatCt++ ;
    } 
  } 
  x_cm = xtot / masstot ; 
  y_cm = ytot / masstot ;
  z_cm = ztot / masstot ;
  
  /* Radius of gyration for _all_ selected atoms */
  rgitot = 0.0 ;
  for (iat = 1 ; iat <= model->number_of_atoms ; iat++) {
    Atom = model->PDB + iat ;
    if (Atom->Selected) {
      rgi = pow((double)(Atom->pdb_xcrd - x_cm),2) +
	pow((double)(Atom->pdb_ycrd - y_cm),2) +
	pow((double)(Atom->pdb_zcrd - z_cm),2);
      rgitot = rgitot + (double)Atom->pdb_masa * rgi ; 
    }
  }
  radius_of_gyration = (float)sqrt(rgitot/(double)masstot);
  
  /*
   * Evaluation of radius of gyration tensor following paper of Garcia de la Torre.
   * They seem to follow a different approach about the atoms taken into account to
   * evaluate the radius of gyration. (further check is needed).
   * Gyration tensor of selected atoms */
  Gxx = Gyy = Gzz = Gxy = Gxz = Gyz = 0.0 ;
  iatCt = 0 ;
  for (iat=1; iat <= model->number_of_atoms; iat++) {
    Atom = model->PDB + iat ;
    if (Atom->Selected) {
      Gxx = Gxx + (Atom->pdb_xcrd-x_cm) * (Atom->pdb_xcrd-x_cm) ;
      Gyy = Gyy + (Atom->pdb_ycrd-y_cm) * (Atom->pdb_ycrd-y_cm) ;
      Gzz = Gzz + (Atom->pdb_zcrd-z_cm) * (Atom->pdb_zcrd-z_cm) ;
      Gxy = Gxy + (Atom->pdb_xcrd-x_cm) * (Atom->pdb_ycrd-y_cm) ;
      Gxz = Gxz + (Atom->pdb_xcrd-x_cm) * (Atom->pdb_zcrd-z_cm) ;
      Gyz = Gxy + (Atom->pdb_ycrd-y_cm) * (Atom->pdb_zcrd-z_cm) ;
      iatCt++ ;
    }
  }
  
  Gyration[0][0] =  Gxx / iatCt ;
  Gyration[0][1] =  Gxy / iatCt ;
  Gyration[0][2] =  Gxz / iatCt ;
  Gyration[1][0] =  Gxy / iatCt ;
  Gyration[1][1] =  Gyy / iatCt ;
  Gyration[1][2] =  Gyz / iatCt ;
  Gyration[2][0] =  Gxz / iatCt ;
  Gyration[2][1] =  Gyz / iatCt ;
  Gyration[2][2] =  Gzz / iatCt ;
  
#ifdef LAPACK
    /* Lapack routines */
    Lower='L';
    dsytrd_(&Lower,&Three1,Gyration,&Three1,EigenValue,OffDiag,TAU,WORK,&LWORK,&ierr);
    if (ierr != 0) fprintf(stderr," Illegal %d\n", ierr);
    dorgtr_(&Lower,&Three1,Gyration,&Three1,TAU,WORK,&LWORK,&ierr);
    if (ierr != 0) fprintf(stderr," Illegal %d\n", ierr);
    Lower='V';
    dsteqr_(&Lower,&Three1,EigenValue,OffDiag,Gyration,&Three2,WORK,&ierr);
    if (ierr != 0) fprintf(stderr," Illegal %d\n", ierr);
    for (iat = 0; iat < 3; iat++) {
        for (jat = 0; jat < 3; jat++) {
              EigenVector[iat][jat] = Gyration[iat][jat] ;
        }
    }
#else  
   /* Two Eispack FORTRAN subroutines are called here. */
   tred2_(&Three1, &Three2, Gyration, EigenValue, Offdiagonal, EigenVector);
   /* EigenVector now contains an orthogonal matrix, hopefully */
  tql2_(&Three1, &Three2, EigenValue, Offdiagonal, EigenVector, &ierr);
  if (ierr != 0)
     printf(" > Eispack: tql2_ returned *NO* convergence after 30 iterations.\n");
#endif 

   Gxx = EigenValue[1] ;
   Gyy = EigenValue[0] ;
   Gzz = EigenValue[2] ;
   G_Rhomb = (Gxx-Gyy)/(Gzz-((Gxx+Gyy)/2.0)) ;

  /*
   * Following TRAMITE paper I am going to compute moment of inertia matrix and 
   * from that the principal moments are computed.
   * Inertia moments */
   EigenValue[0] = EigenValue[1] = EigenValue[2] = 0.0 ;
   Ixx = Iyy = Izz = Ixy = Ixz = Iyz = 0.0 ; 
   for(iat = 1 ; iat <= model->number_of_atoms ; iat++) {
     Atom = model->PDB + iat ;
     if ( Atom->Selected ) {
       Ixx = Ixx + Atom->pdb_masa * (pow(Atom->pdb_ycrd-y_cm,2) + pow(Atom->pdb_zcrd-z_cm,2));
       Iyy = Iyy + Atom->pdb_masa * (pow(Atom->pdb_xcrd-x_cm,2) + pow(Atom->pdb_zcrd-z_cm,2));
       Izz = Izz + Atom->pdb_masa * (pow(Atom->pdb_xcrd-x_cm,2) + pow(Atom->pdb_ycrd-y_cm,2));
       Ixy = Ixy + Atom->pdb_masa * Atom->pdb_xcrd-x_cm * Atom->pdb_ycrd-y_cm;
       Ixz = Ixz + Atom->pdb_masa * Atom->pdb_xcrd-x_cm * Atom->pdb_zcrd-z_cm;
       Iyz = Iyz + Atom->pdb_masa * Atom->pdb_ycrd-y_cm * Atom->pdb_zcrd-z_cm;
     }
   }
   
   /* Storage of the computed moments of inertia, units are discarded 
    *   1.0e-7 to convert from g cm2 to Kg m2 
    *   NA (Avogadro) 
    *   Angstroms -> meters
    */
   model->Ixx = Ixx ;
   model->Iyy = Iyy ;
   model->Izz = Izz ;
   model->Ixy = Ixy ;
   model->Ixz = Ixz ;
   model->Iyz = Iyz ;
  
   /* Prepare to obtain principal values by computing eigen values */
   Inertia[0][0] =  model->Ixx ;
   Inertia[0][1] = -model->Ixy ;
   Inertia[0][2] = -model->Ixz ;
   Inertia[1][0] = -model->Ixy ;
   Inertia[1][1] =  model->Iyy ;
   Inertia[1][2] = -model->Iyz ;
   Inertia[2][0] = -model->Ixz ;
   Inertia[2][1] = -model->Iyz ;
   Inertia[2][2] =  model->Izz ;
  
#ifdef LAPACK
    /* Lapack routines */
    Lower='L';
    dsytrd_(&Lower,&Three1,Inertia,&Three1,EigenValue,OffDiag,TAU,WORK,&LWORK,&ierr);
    if (ierr != 0) fprintf(stderr," Illegal %d\n", ierr);
    dorgtr_(&Lower,&Three1,Inertia,&Three1,TAU,WORK,&LWORK,&ierr);
    if (ierr != 0) fprintf(stderr," Illegal %d\n", ierr);
    Lower='V';
    dsteqr_(&Lower,&Three1,EigenValue,OffDiag,Inertia,&Three2,WORK,&ierr);
    if (ierr != 0) fprintf(stderr," Illegal %d\n", ierr);
    for (iat = 0; iat < 3; iat++) {
        for (jat = 0; jat < 3; jat++) {
              EigenVector[iat][jat] = Inertia[iat][jat] ;
        }
    }
#else 
    /* Eispack routines, we have the code */
    tred2_(&Three1, &Three2, Inertia, EigenValue, Offdiagonal, EigenVector);
    tql2_(&Three1, &Three2, EigenValue, Offdiagonal, EigenVector, &ierr);
    if (ierr != 0)
       printf(" > optimal_rotate: tql2_ returned *NO* convergence after 30 iterations.\n");
#endif  

   /* TRAMITE procedure to make an analog of second rank order tensor
    * from inertia principal moments. The eigenvalues are procssed and
    * selected to fulfill traceless condition 
    * NOTE: See TRAMITE paper.
    */
   A11 = EigenValue[0] - ( EigenValue[1] + EigenValue[2] ) / 2.0 ;
   A22 = EigenValue[1] - ( EigenValue[0] + EigenValue[2] ) / 2.0 ;
   A33 = EigenValue[2] - ( EigenValue[0] + EigenValue[1] ) / 2.0 ;

   if ( -A22 > 0.0 ) {
	   Axx = A22 ;
	   Ayy = A11 ;
	   Azz = A33 ;
   	   for (iat=0;iat<3;iat++) Xaxe[iat] = EigenVector[1][iat];
           for (iat=0;iat<3;iat++) Yaxe[iat] = EigenVector[0][iat];
           for (iat=0;iat<3;iat++) Zaxe[iat] = EigenVector[2][iat];
   } else {
	   Axx = A22 ;
	   Ayy = A33 ;
	   Azz = A11 ;
   	   for (iat=0;iat<3;iat++) Xaxe[iat] = EigenVector[1][iat];
           for (iat=0;iat<3;iat++) Yaxe[iat] = EigenVector[2][iat];
           for (iat=0;iat<3;iat++) Zaxe[iat] = EigenVector[0][iat];
   }

   /* Rhombicity evaluated after the procedure */
   I_Rhomb = (Axx-Ayy)/(Azz-((Axx+Ayy)/2.0)) ;

   printf("\n> Rgyration: ---Model %3d---\n", model->modelno) ;
   printf("> Rgyration: Center of Masses: x_cm(%.3f), y_cm(%.3f), z_cm(%.3f) \n", x_cm, y_cm, z_cm);
   printf("> Rgyration: Radius of Gyration <S2> :  \n");
   printf("                Computed: %6.3f    Predicted (2.2N^0.38): %6.3f \n", 
	  radius_of_gyration, 2.2*pow(model->number_of_residues,0.38));
   printf("> Rgyration: EigenValues of Gyration tensor: \n");
   printf("                Gxx: %10.1lf  Gyy: %10.1lf  Gzz: %10.1lf R: %5.3lf\n", Gxx, Gyy, Gzz, G_Rhomb);
   /*
    * printf("> Rgyration: Inertia moments: [kg. m2]\n");
    * printf("> Rgyration:    Ixx: %10.3le  Iyy: %10.3le  Izz: %10.3le \n", model->Ixx, model->Iyy, model->Izz);
    * printf("> Rgyration:    Ixy: %10.3le  Ixz: %10.3le  Iyz: %10.3le \n", model->Ixy, model->Ixz, model->Iyz);
    */
   printf("> Inertia  : Principal Inertia Moments [EigenValues]: \n");
   printf("                Ixx: %10.3le  Iyy: %10.3le  Izz: %10.3le \n", EigenValue[0], EigenValue[1], EigenValue[2]);
   printf(">               Axx: %10.3le  Ayy: %10.3le  Azz: %10.3le R: %5.3lf .. %5.3lf\n", Axx, Ayy, Azz, I_Rhomb,2.0*(Axx-Ayy)/(3.0*Azz));
   printf("> Inertia  : Axes of alignment (EigenVectors): \n");
   printf(">               Axx: %10.3le  Axy: %10.3le  Axz: %10.3le \n", Xaxe[0], Xaxe[1], Xaxe[2]);
   printf(">               Ayx: %10.3le  Ayy: %10.3le  Ayz: %10.3le \n", Yaxe[0], Yaxe[1], Yaxe[2]);
   printf(">               Azx: %10.3le  Azy: %10.3le  Azz: %10.3le \n", Zaxe[0], Zaxe[1], Zaxe[2]);

   return ;
}

