/*$Id: phi_psi_chi.c,v 3.32 2003/10/06 12:13:32 roberto Exp roberto $*/

#include "pdbstat.h"

/* ========================================================================= */
void phi_psi_chi(FILE *rtti,int choice)
{
/*
  ---
  --- This is the driver to calculate the dihedrals phi, psi and chi for a
  --- molecule using the routine dihang. Here the atoms are picked up if 
  --- they form any of the dihedral. The algoritm to pick up the atoms is
  --- too primitive and crude, it takes a VERY LONG time to evaluate the
  --- dihedrals even for a little molecule.
  ---
  --- On May 20, 1993 the algorithm has been improved a little by adding the 
  --- first 'if' which speeds up the process (RTT, May 20, 1993, CABM). On the
  --- same day the process has been finally completely speeded up with three
  --- more 'if'
  --- 
  --- Now (May, 28) it is able to calculate also the omega and planarity
  --- about the peptide.
  ---
  --- On June, 9 added the possibility to check quality of protein calculating
  --- the special dihedral CA-N-C-CB. It should have a value of 33.8 +/- 4.2
  --- which is the standard value for an L-aminoacid as derived from 93,000
  --- residues studied (see reference).
  --- Ref: A. Morris et al., Prot. Struct. Func. and Gen., 12:345-364(1992).
  ---
  --- (Roberto Tejero, RTT, May 1993, CABM-Rutgers University).
  --- 
  --- On JAN. 1994 the code was deeply changed to fully optimize (I hope)
  --- and check the calculation. Now the file is sensibly shorter because
  --- the code is less (and more efficient!)
  ---
  --- Very important NOTE:
  ---   I discovered that we HAVE to have the coordinates with the FOLLOWING
  ---   order:   N, CA, side chain,  C, O      or
  ---            N, side chain, CA , C, O   but no other orders.
  ---   as the algorithm is HIGHLY dependent on order in atom types (it assumes
  ---   that usually you HAVE : N, CA, side chain, C, O.
  ---    
  --- (Roberto Tejero, RTT, Jan 1994, Valencia University).
  ---
*/
     static int   i, k ;
     static int   tmp, tempvar,numi,numj,numk,numl, EndAtom;
     static int   count_phi, count_psi, count_chi1, count_res, count_omega;
     static int   count_planaro, count_planarh, count_chiral, count_chiral_sc ;
     static char  angulos[MAXRESIDUES][80], char_tmp[10];
     static char  t[3];
     
     static struct models *model ;
     
     /*
      * Initialize some variables for calculation
      */
     count_phi=1;
     count_psi=0;
     count_chi1=0;
     count_omega=0;
     count_planaro=0;
     count_planarh=1;
     count_chiral=0;
     count_chiral_sc=0;
     count_res = 1 ; 
     
     model = MODELS+key_model ;
     
     MAX = model->number_of_atoms ;
     
     /*
      * Check if we have *REALLY* some molecule in memory
      */
     if (MAX <= 0) {
	  printf("> PHI_PSI_CHI: * No molecule read \n");
	  return ;
     }
     
     tmp = model->PDB[1].pdb_num_res ;
     
     /*
      * If ALA we have NO CHI1
      */
     if (strncmp(model->DIH[count_res].resname,"ALA",3)==0) {
	  count_chi1++;
     }
     
     /*
      * If GLY we have NO CHI1 and NO chiral
      */
     if (strncmp(model->DIH[count_res].resname,"GLY",3)==0)  {
	  count_chi1++;
	  count_chiral++;
     }
     
     /*
      * If PRO we have NO planar
      */
     if (strncmp(model->DIH[count_res].resname,"PRO",3)==0) {
	  count_planarh++;
     }
     
     /*
      * If ILE or THR, count side chain chirality
      * if NOT ile or thr, we have NO side chain chirality
      */
     if ((strncmp(model->DIH[count_res].resname,"ILE",3)==0) ||
	 (strncmp(model->DIH[count_res].resname,"THR",3)==0)) {
	  ;
     } else {
	  count_chiral_sc++;
     }
     
     tempvar = (int) (model->number_of_residues / 2);
     
     /* Main loop through all atoms */
     for (i = 1 ; i < MAX ; i++) {
	  numi = numj = numk = numl = EndAtom = 0 ;
	  EndAtom = i + 40 ;
	  if (EndAtom >= MAX) EndAtom = MAX ;
	  /* 
	   * watch for residues with no angles 
	   */			
	  if (tmp != model->PDB[i].pdb_num_res) {
	       count_res++;
	       tmp = model->PDB[i].pdb_num_res;
	       if (strncmp(model->DIH[count_res].resname,"ALA",3)==0) {
		    count_chi1++;
	       }
	       if (strncmp(model->DIH[count_res].resname,"GLY",3)==0) {
		    count_chi1++;
		    count_chiral++;
	       }
	       if (strncmp(model->DIH[count_res].resname,"PRO",3)==0) {
		    count_planarh++;
	       }
	       if ((strncmp(model->DIH[count_res].resname,"THR",3)==0) ||
		   (strncmp(model->DIH[count_res].resname,"ILE",3)==0)) {
		    ;
	       } else {
		    count_chiral_sc++;
	       }
	  }
	  
	  /*   residues i+1=j=k=l ,   angle: phi C-N-CA-C */
	  numi = numj = numk = numl = 0 ;
	  for (k = i+1; k < EndAtom ; k++) {      
	       if ((strncmp(model->PDB[i].pdb_iupac," C  ",4)==0) &&
		   (model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res))
		    numi = i ;
	       if (((strncmp(model->PDB[k].pdb_iupac," N  ",4)==0) ||
		    (strncmp(model->PDB[k].pdb_iupac," NT ",4)==0) ||
		    (strncmp(model->PDB[k].pdb_iupac," NP ",4)==0)) &&
		   (model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res))
		    numj = k ;
	       if ((strncmp(model->PDB[k].pdb_iupac," CA ",4)==0) &&
		   (model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res))
		    numk = k ;
	       if ((strncmp(model->PDB[k].pdb_iupac," C  ",4)==0) &&
		   (model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res))
		    numl = k ;
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_phi++;
	       if (count_phi > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval: * Error in phi \n");
		    exit(-1);
	       }
	       model->DIH[count_res+1].phi = dihang(numi,numj,numk,numl)/radperdeg;
	  }
	  
	  /*   residues i+1=j=k=l ,   angle: planar_h (Planarity of hydrogen) */
	  numi = numj = numk = numl = 0 ;
	  for (k = i+1; k < EndAtom ; k++) {      
	       if (model->PDB[i].pdb_num_res + 1 == model->PDB[k].pdb_num_res) {
		    if (strncmp(model->PDB[i].pdb_iupac," C  ",4)==0)  numi = i ;
		    if ((strncmp(model->PDB[k].pdb_iupac," N  ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," NT ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," NP ",4)==0))  
			 numj = k ;
		    if (strncmp(model->PDB[k].pdb_iupac," CA ",4)==0) numk = k ;
		    if ((strncmp(model->PDB[k].pdb_iupac," H  ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," HN ",4)==0))
			 numl = k ;
	       }
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_planarh++;
	       if (count_planarh > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval:* Error in planar h\n");
		    exit(-1);
	       }
	       model->DIH[count_res+1].planar_h = dihang(numj,numi,numl,numk)/radperdeg;
	  }
	  
	  /* residue i=j=k=l-1 , angle: psi  N-CA-C-N */
	  numi = numj = numk = numl = 0 ;
	  for ( k = i+1 ; k < EndAtom ; k++) {
	       if  (model->PDB[i].pdb_num_res == model->PDB[k].pdb_num_res) {
		    if ((strncmp(model->PDB[i].pdb_iupac," N  ",4)==0) ||
			(strncmp(model->PDB[i].pdb_iupac," NT ",4)==0) ||
			(strncmp(model->PDB[i].pdb_iupac," NP ",4)==0)) 
			 numi = i ;
		    if (strncmp(model->PDB[k].pdb_iupac," CA ",4)==0) numj = k ;
		    if (strncmp(model->PDB[k].pdb_iupac," C  ",4)==0) numk = k ;
	       }
	       if  (model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res) {
		    if ((strncmp(model->PDB[k].pdb_iupac," N  ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," NT ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," NP ",4)==0))
			 numl = k ; 
	       }
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_psi++;
	       if (count_psi > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval: * Error in psi \n");
		    exit(-1);
	       }
	       model->DIH[count_res].psi = dihang(numi,numj,numk,numl)/radperdeg;
	  }
	  
	  /* residue i=j=k=l-1 , angle: planar_o  (Planarity of oxigen) */    
	  numi = numj = numk = numl = 0 ;
	  for ( k = i+1 ; k < EndAtom ; k++) {
	       if (model->PDB[i].pdb_num_res == model->PDB[k].pdb_num_res) {
		    if ((strncmp(model->PDB[i].pdb_iupac," CA ",4)==0)) numi = i ;
		    if (strncmp(model->PDB[k].pdb_iupac," C  ",4)==0)   numj = k ;
		    if (strncmp(model->PDB[k].pdb_iupac," O  ",4)==0)   numk = k ;
	       }
	       if (model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res) {
		    if ((strncmp(model->PDB[k].pdb_iupac," N  ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," NT ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," NP ",4)==0))
			 numl = k ;
	       }
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_planaro++;
	       if (count_planaro > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval: * Error in planar oxygen \n");
		    exit(-1);
	       }
	       model->DIH[count_res].planar_o= dihang(numj,numi,numl,numk)/radperdeg;
	  }
	  
	  /* residues: i=j=k+1=l+1 , angles: omega CA-C-N-CA */
	  numi = numj = numk = numl = 0 ;
	  for ( k = i+1 ; k < EndAtom ; k++) {
	       if (model->PDB[i].pdb_num_res == model->PDB[k].pdb_num_res) {
		    if (strncmp(model->PDB[i].pdb_iupac," CA ",4)==0)  numi = i ;
		    if (strncmp(model->PDB[k].pdb_iupac," C  ",4)==0)  numj = k ;
	       }
	       if (model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res) {
		    if (((strncmp(model->PDB[k].pdb_iupac," N  ",4)==0) ||
			 (strncmp(model->PDB[k].pdb_iupac," NT ",4)==0) ||
			 (strncmp(model->PDB[k].pdb_iupac," NP ",4)==0)) &&
			(model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res))
			 numk = k ;
		    if ((strncmp(model->PDB[k].pdb_iupac," CA ",4)==0) &&
			(model->PDB[i].pdb_num_res+1 == model->PDB[k].pdb_num_res))
			 numl = k ;
	       }
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_omega++;
	       if (count_omega > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval: *** Error in omega \n");
		    exit(-1);
	       }
	       model->DIH[count_res].omega = dihang(numi,numj,numk,numl)/radperdeg;
	  }
	  
	  /* residues: i=j=k=l , angles: chi1 N-CA-CB-*G */
	  numi = numj = numk = numl = 0 ;
	  for ( k = i+1 ; k < EndAtom ; k++) {
	       if (model->PDB[i].pdb_num_res == model->PDB[k].pdb_num_res) {
		    if ((strncmp(model->PDB[i].pdb_iupac," N  ",4)==0) ||
			(strncmp(model->PDB[i].pdb_iupac," NT ",4)==0) ||
			(strncmp(model->PDB[i].pdb_iupac," NP ",4)==0))
			 numi = i ;
		    if (strncmp(model->PDB[k].pdb_iupac," CA ",4)==0)
			 numj = k ;
		    if (strncmp(model->PDB[k].pdb_iupac," CB ",4)==0)
			 numk = k ;
		    if ((strncmp(model->PDB[k].pdb_iupac," CG1",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," CG ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," OG1",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," OG ",4)==0) ||
			(strncmp(model->PDB[k].pdb_iupac," SG ",4)==0)) 
			 numl = k ;
	       }
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_chi1++;
	       if (count_chi1 > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval: *** Error in chi \n");
		    exit(-1);
	       }
	       model->DIH[count_res].chi1 = dihang(numi,numj,numk,numl)/radperdeg;
	  }
	  
	  /*  residues: i=j=k=l , --- chiral CA-N-C-CB    --- */
	  numi = numj = numk = numl = 0 ;
	  for ( k  = i+1 ; k < EndAtom ; k++) {
	       if (model->PDB[i].pdb_num_res == model->PDB[k].pdb_num_res) {
		    if (strncmp(model->PDB[k].pdb_iupac," CA ",4)==0) numj = k ;
		    if (strncmp(model->PDB[k].pdb_iupac," C  ",4)==0) numk = k ;
		    if ((strncmp(model->PDB[i].pdb_iupac," N  ",4)==0) ||
			(strncmp(model->PDB[i].pdb_iupac," NT ",4)==0) ||
			(strncmp(model->PDB[i].pdb_iupac," NP ",4)==0))
			 numi = i ;
		    if (strncmp(model->PDB[k].pdb_iupac," CB ",4)==0) numl = k ;
	       }
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_chiral++;
	       if (count_chiral > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval: * Error in chiral \n");
		    exit(-1);
	       }
	       model->DIH[count_res].chiral = dihang(numj,numi,numk,numl)/radperdeg;
	  }
	  
	  /*  residues: i=j=k=l , --- side chain chirality for ILE and THR    */
	  numi = numj = numk = numl = 0 ;
	  for ( k = i+1 ; k < EndAtom ; k++) {
	       if (model->PDB[i].pdb_num_res == model->PDB[k].pdb_num_res) 
		    if ((strncmp(model->PDB[i].pdb_res,"THR",2)==0) ||
			(strncmp(model->PDB[i].pdb_res,"ILE",2)==0)) {
			 if (strncmp(model->PDB[i].pdb_iupac," CA ",4)==0) 
			      numi = i ;
			 if (strncmp(model->PDB[k].pdb_iupac," CB ",4)==0)
			      numj = k ;
			 
			 if ((strncmp(model->PDB[k].pdb_iupac," CG1",4)==0) &&
			     (strncmp(model->PDB[k].pdb_res,"ILE",2)==0))
			      numk = k ;
			 
			 if ((strncmp(model->PDB[k].pdb_iupac," CG2",4)==0) &&
			     (strncmp(model->PDB[k].pdb_res,"ILE",2)==0))
			      numl = k ;
			 
			 if ((strncmp(model->PDB[k].pdb_iupac," OG1",4)==0) &&
			     (strncmp(model->PDB[k].pdb_res,"THR",2)==0))
			      numk = k ;	  
			 
			 if ((strncmp(model->PDB[k].pdb_iupac," CG2",4)==0) &&
			     (strncmp(model->PDB[k].pdb_res,"THR",2)==0))
			      numl = k ;
		    }
	  }
	  if (numi !=0 && numj !=0 && numk !=0 && numl !=0) {
	       count_chiral_sc++;
	       if (count_chiral_sc > MAXRESIDUES) {
		    printf("> PHI_PSI_CHI_eval: * Error in side chain chiral \n");
		    exit(-1);
	       }
	       model->DIH[count_res].chiral_sc =  dihang(numj,numk,numl,numi)/radperdeg;
	  }
     }
     
     /*
      * Printing results (sreen and file)
      */
     char_tmp[1]='\0';
     sprintf(char_tmp,"\n");
     /* Initialize */
     for (i=0;i<MAXRESIDUES;i++)
	  for (k=0;k<80;k++) {
	       model->angulos[i][k]='\0';
	       angulos[i][k]='\0';
	  }
     /* Write angles */
     for (i = 1 ; i <= model->number_of_residues;  i++) {
	  strcpy(angulos[i],"");  
	  sprintf(char_tmp," %s %2s %3d", model->DIH[i].resname, model->DIH[i].chain_id, model->DIH[i].resnumber);
	  strcat(angulos[i],char_tmp);
	  strcat(model->angulos[i],char_tmp);
	  
	  /*--- PHI ---*/
	  if (model->DIH[i].phi != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].phi);
	  else 
	       sprintf(char_tmp,"        ");
	  strcat(angulos[i],char_tmp);
	  strcat(model->angulos[i],char_tmp);
	  
	  /*--- PSI ---*/
	  if (model->DIH[i].psi != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].psi);
	  else 
	       sprintf(char_tmp,"        ");
	  strcat(angulos[i],char_tmp);
	  strcat(model->angulos[i],char_tmp);
	  
	  /*---OMEGA---*/
	  if (model->DIH[i].omega != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].omega); 
	  else 
	       sprintf(char_tmp,"        ");
	  strcat(angulos[i],char_tmp);
	  strcat(model->angulos[i],char_tmp);
	  
	  /*---CHI1---*/
	  if (model->DIH[i].chi1 != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].chi1);
	  else 
	       sprintf(char_tmp,"        ");
	  strcat(angulos[i],char_tmp);
	  strcat(model->angulos[i],char_tmp);
	  
	  /*---PLANARTY---*/
	  if (model->DIH[i].planar_o != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].planar_o);
	  else 
	       sprintf(char_tmp,"        ");
	  strcat(angulos[i],char_tmp);
	  if (model->DIH[i].planar_h != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].planar_h);
	  else 
	       sprintf(char_tmp,"        ");
	  strcat(angulos[i],char_tmp);
	  
	  /*---CHIRAL---*/
	  if (model->DIH[i].chiral != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].chiral);
	  else 
	       sprintf(char_tmp,"        ");
	  strcat(angulos[i],char_tmp);
	  
	  /*---CHIRAL side chain ---*/
	  if (model->DIH[i].chiral_sc != 11111.0) 
	       sprintf(char_tmp,"%8.2f",model->DIH[i].chiral_sc);
	  else 
	       sprintf(char_tmp,"        ");  
	  strcat(angulos[i],char_tmp);
     }
     
     /*
      * Writing results to File
      */
     if ((choice == 1) || (choice == 999)) {
	  if (choice == 999) fprintf(rtti,"\n\n\n");
	  fprintf(rtti,"\n  PHI PSI CHI for molecule: `%s', model #%d (%2d) [E:%7.3f]\n ",
		  name_of_the_file_coords, 
		  key_model, 
		  model->modelno,
		  model->energy);
	  fprintf(rtti,"\n                              180.0           0.0     0.0"\
		  "    29-38    sidech");
	  fprintf(rtti,"\n   RESIDUE     PHI     PSI    OMEGA   CHI1   PLAN_O  PLAN_H  CHIRAL   CHIRAL\n");
	  fprintf(rtti," __________________________________________________________________________ \n");
     }
     
     for (i = 1 ; i <= model->number_of_residues;  i++)
	  if ((choice == 1) || (choice == 999))
	       fprintf(rtti,"\n %s",angulos[i]);
     
     /*
      * Writing results to  Screen. First clean up the screen adding 50 blank 
      * lines and then print the results in groups of 20 lines
      */
     if (choice == 1) 
	  for (i = 1 ; i <= 50 ; i++)
	       printf("\n");
     
     if ((choice == 1) || (choice == 999)) {
	  if (choice == 999) printf("\n\n\n");
	  printf("\n  PHI PSI CHI for molecule: `%s', model #%d(%2d) [E:%7.3f]\n ",
		 name_of_the_file_coords, key_model, model->modelno, model->energy);
	  printf("\n                              180.0           0.0     0.0"\
		 "    29-38    sidech");
	  printf("\n   RESIDUE     PHI     PSI    OMEGA   CHI1   PLAN_O  PLAN_H"\
		 "  CHIRAL   CHIRAL\n");
	  printf(" ____________________________________________________________"\
		 "______________ \n");
     }
     
     for (i = 1 ; i <= model->number_of_residues;  i++)
	  if ((choice == 1) || (choice == 999)) {
	       if (choice == 1)
		    if (fmod(i,22) == 0) {
			 printf("\n> ... Hit return when ready ... or 'q'+return to exit");
			 GetFromKeyBoard(t, MAXWORD, stdin);
			 if ( (strncasecmp(t,"q",1)==0) || (strncasecmp(t,"e",1)==0) )
			      return ;
		    }
	       printf(" %s\n",angulos[i]);
	  }
     
     /*
      * In case we have done already the calculation skip up to here
      * making nothing. Nada is nothing in spanish, so it's a blank line
      */
     
     return;
}
