package Bmrb;

use Carp;

# Perl library to parse the sequence and assigned chemical shifts from a BMRB file
#
# created 10/21/09 - Gautam Singh
# modified 10/30/09 - corrects ambiguity code mismatch between carbon and corresponding hydryogen atoms

sub new {
	my $class = shift;
	my $self = {};
	$self->{filename} = shift;
	$self->{filepath} = shift;
	$self->{debug} = shift;
	$self->{one_letter_sequence} = "";
	$self->{residue_count} = 0;
	$self->{monomeric_polymer_frame} = [];
	$self->{assigned_chemical_shifts_frame} = [];
	$self->{extracted_chemical_shifts} = {};
	$self->{residues} = {};
	$self->{version} = undef;
	$self->{pseudoatom} = {
		'GLYQA' => 'QA',
		'ALAMB' => 'HB',
		'VALMG1' => 'HG1',
		'VALMG2' => 'MG2',
		'ILEMG' => 'HG',
		'ILEMD' => 'HD',
		'LEUMD1' => 'HD1',
		'LEUMD2' => 'HD2',
		'THRMG' => 'HG',
		'LYSQZ' => 'HZ',
		'METME' => 'HE'
	};
	bless($self, $class);
	$self->parse_bmrb();
	if ($self->{version} eq "autostructure") {

	} else {
		$self->build_sequence();
	}
	$self->parse_chem_shifts();
	return $self;
}
sub extract_sequence() {
	my $self = shift;
	my $i = 0;
	my $output = "";
	foreach my $key (sort{$a<=>$b}(keys (%{$self->{residues}}))) {
		$output .= $key . "\t@\t" . $self->{residues}->{$key};
		$i++;
		if ($i >= 5) {
			$output .= "\n";
			$i = 0;
		} else {
			$output .= "\t";
		}
	}
	$output .= "\n";
	return $output;
}
sub build_sequence {
	my $self = shift;
	my $lines = $self->{monomeric_polymer_frame};
	my $i = 0;
	my $line;
	for ($line = @$lines[$i]; $i < scalar(@$lines); $i++, $line = @$lines[$i]) {
		if ($line =~ /_mol_residue_sequence/i) {
			if (@$lines[++$i] =~ /;/) {
				while (@$lines[++$i] !~ /;/) {
					chomp($line = @$lines[$i]);
					$self->{one_letter_sequence} .= $line;
				}
				$self->{residue_count} = length($self->{one_letter_sequence});
			}
		}
		if ($line =~ /loop_/) {
			my $loop = {};
			my $loop_headers = {};
			my $header_count = 0;
			my $seq_code_col = 0;
			my $label_col = 0;
			my $all_headers = "";
			while (@$lines[++$i] !~ /stop_/) {
				next if (@$lines[$i] =~ /^#/ or @$lines[$i] =~ /^\s*$/);
				while (($line = @$lines[$i]) =~ /^\s*(_\S+)/) {
					next if (@$lines[$i] =~ /^#/ or @$lines[$i] =~ /^\s*$/);
					$header_count++;
					$all_headers .= lc($1);
					if ($line =~ /residue_label/i or $line =~ /_Entity_comp_index\.Comp_ID/i) {
						$loop_headers{'label'} = $header_count;
					}
					if ($line =~ /residue_seq_code/i or $line =~ /_Entity_comp_index\.ID/i) {
						$loop_headers{'seq_code'} = $header_count;
					}
					$i++;
				}
				if (($all_headers =~ /residue_seq_code/ and $all_headers =~ /_residue_label/) or ($all_headers =~ /entity_comp_index\.comp_id/ and $all_headers =~ /entity_comp_index\.id/)) {
					my @items = split(/\s+/, $line);
					$self->debug("seq_code col is " . $loop_headers{'seq_code'});
					$self->debug("label col is " . $loop_headers{'label'});
					for (my $j = 1; $j <= scalar(@items); $j++) {
						if (($j % $header_count) == ($loop_headers{'seq_code'} % $header_count)) {
							$seq_code = $items[$j - 1];
						}
						if (($j % $header_count) == ($loop_headers{'label'}) % $header_count) {
							$label = $items[$j - 1];
						}
						if ($j % $header_count == 0) {
							$self->{residues}{0 + $seq_code} = $label;
							$self->debug("adding: $seq_code -> $label");
						}
					}
				}
			}
		}
	}
}
sub parse_chem_shifts {
	my $self = shift;
	my $temp_line;
	my $inside_chem_shifts = 0;
	my $output = "";
	my @stk = undef;
	my $chem_shifts_ref = $self->{extracted_chemical_shifts};
	my %chem_shifts = %$chem_shifts_ref;
	my $lines = $self->{assigned_chemical_shifts_frame};
	my $k = 0;
	for (my $i = 0; $i < scalar(@$lines); $i++) {
		$temp_line = @$lines[$i];
		if ($temp_line =~ m/loop_/) {
			my @headers = [];
			my $header_col = 0;
			my ($id_col, $authcode_col, $seqcode_col, $label_col, $name_col, $type_col, $value_col, $error_col, $ambig_col);
			while (($i < scalar(@$lines)) and (($temp_line = @$lines[++$i]) =~ /^\s*_/)) {
				$header_col++;
				$self->debug($temp_line);
				push(@headers, $temp_line);
				if ($temp_line =~ /shift_assign_id/i or $temp_line =~ /Atom_chem_shift\.ID/i) {
					$id_col = $header_col;
				}
				if ($temp_line =~ /author_seq_code/i or $temp_line =~ /_Atom_chem_shift\.Auth_seq_ID/i) {	#residue #
					$authcode_col = $header_col;
				}
				if ($temp_line =~ /seq_code/i or $temp_line =~ /_Atom_chem_shift\.Comp_index_ID/i) {	#residue #
					$seqcode_col = $header_col;
				}
				if ($temp_line =~ /label/i or $temp_line =~ /_Atom_chem_shift\.Comp_ID/i) {	#3 letter code
					$label_col = $header_col;
				}
				if ($temp_line =~ /atom_name/i or $temp_line =~ /_Atom_chem_shift\.Atom_ID/i) {	#atom name eg CA, HG1
					$name_col = $header_col;
				}
				if ($temp_line =~ /atom_type/i or $temp_line =~ /_Atom_chem_shift\.Atom_type/i) {	#atom type eg, C, H
					$type_col = $header_col;
				}
				if ($temp_line =~ /_Chem_shift_value\s*$/i or $temp_line =~ /_Atom_chem_shift\.Chem_shift_val\s*$/i or $temp_line =~ /_Atom_chem_shift\.Val\s*$/i) {
					$self->debug("inside chem shifts");
					$value_col = $header_col;
					$inside_chem_shifts = 1;
				}
				if ($temp_line =~ /shift_value_error/i or $temp_line =~ /_Atom_chem_shift\.Chem_shift_val_err/i or $temp_line =~ /_Atom_chem_shift\.Val_err\s*$/i) {
					$error_col = $header_col;
				}
				if ($temp_line =~ /ambiguity/i or $temp_line =~ /_Atom_chem_shift\.Chem_shift_ambiguity_code/i) {
					$ambig_col = $header_col;
				}
			}
			if ($inside_chem_shifts) {
				my ($prev_id, $prev_seqcode, $prev_label, $prev_name, $prev_type, $_prev_value, $prev_error, $prev_ambig);
				my $methyl_H_counter = 1;
				my $methyl_H_type = "";
				while (($i < scalar(@$lines)) and not $temp_line =~ /stop_/) {
					#TODO... replace this with split()
					if ($temp_line =~ /\s*([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)\s*([^\s]*)/i) {
						my ($id, $seqcode, $label, $name, $type, $value, $error, $ambig) = ($$id_col, $$seqcode_col, $$label_col, $$name_col, $$type_col, $$value_col, $$error_col, $$ambig_col);
						if (exists $self->{pseudoatom}->{uc($label . $name)}) {
							#$name .= $self->{pseudoatom}->{uc($label . $name)};
						}
						if ($self->{residues}->{$seqcode} =~ $label) {
							unless (defined $chem_shifts{$seqcode}) {
								$chem_shifts{$seqcode} = {};
							}
							my $assigned_atoms_ref = $chem_shifts{$seqcode};
							$assigned_atoms_ref->{$name} = {id => "$id", label => "$label", name => uc("$name"), type => "$type", value => "$value", error => "$error", ambig => "$ambig"};
							$k++;
							$self->debug("adding: $id $seqcode $label $name $type $value $error $ambig");
						}
						$prev_name = $name;
						$prev_seqcode = $seqcode;
					}
					$temp_line = @$lines[++$i];
				}
			}
		}
	}
	$self->{extracted_chemical_shifts} = \%chem_shifts;
	#return $self->correct_bmrb();
}
sub extract_bmrb {
	my $self = shift;
	my @stk = undef;
	my $chem_shifts_ref = $self->{extracted_chemical_shifts};
	my %chem_shifts = %$chem_shifts_ref;
	$self->debug("seqcode size is : " . scalar(keys (%chem_shifts)));
	foreach my $seqcode (keys (%chem_shifts)) {
		my $residue_assignment = $chem_shifts{$seqcode};
		$self->debug("seqcode: $seqcode \tshift counts are : " . scalar(keys (%$residue_assignment)));
		foreach my $atom_name (keys (%$residue_assignment)) {
			unless (exists $residue_assignment->{$atom_name}) {
				next;
			}
			my $atom_assignment = $residue_assignment->{$atom_name};
			my $id = $atom_assignment->{'id'};
			my $name = $atom_assignment->{'name'};
			my $type = $atom_assignment->{'type'};
			my $value = $atom_assignment->{'value'};
			my $error = $atom_assignment->{'error'};
			my $ambig =  $atom_assignment->{'ambig'};
			if ((defined $name) and $name =~ /([a-zA-Z0-9]+)(3)/i) {
				#print "$id: #$name#: $1: " . $2 . "\n";
				if (defined $residue_assignment->{$1 . "1"} and defined $residue_assignment->{$1 . "2"}) {
					$self->debug("\tdeleting: " . $residue_assignment->{$1 . "1"}->{'id'} . " and " . $residue_assignment->{$1 . "2"}->{'id'});
					my %new_atom_assignment = %$atom_assignment;
					#$residue_assignment->{$1 . "1"} = undef;
					#$residue_assignment->{$1 . "2"} = undef;
					#$residue_assignment->{$1 . "3"} = undef;
					delete ($residue_assignment->{$1 . "1"});
					delete ($residue_assignment->{$1 . "2"});
					delete ($residue_assignment->{$1 . "3"});
					$residue_assignment->{$1} = \%new_atom_assignment;
					$residue_assignment->{$1}->{'name'} = $1;
					#print "$seqcode: id $id added new $1: " . $residue_assignment->{$1}->{'name'} . " ::: " . $residue_assignment->{$1}->{'ambig'} . "\n";
				}
			}
		}
	}
	foreach my $seqcode (keys (%chem_shifts)) {
		my $residue_assignment = $chem_shifts{$seqcode};
		foreach my $atom_assignment (values (%$residue_assignment)) {
			my $id = $atom_assignment->{'id'};
			my $name = $atom_assignment->{'name'};
			my $type = $atom_assignment->{'type'};
			my $value = $atom_assignment->{'value'};
			my $error = $atom_assignment->{'error'};
			my $ambig =  $atom_assignment->{'ambig'};
			if ((defined $atom_assignment->{'label'}) and $atom_assignment->{'label'} =~ /(leu)/i) {
				if ($name =~ /(cd)(\d)$/i) {
					my $ambig = $atom_assignment->{'ambig'};
					foreach my $key (%$residue_assignment) {
						#print "key: $key and looking for: HD$2\n";
					}
					if (exists $residue_assignment->{"HD" . $2}) {
						my $methyl_proton_assignment = $residue_assignment->{"HD" . $1};
						$residue_assignment->{"HD" . $2}->{'ambig'} = $ambig;
						$self->debug("setting ambig code at $id for HD$2 to $ambig");
					} else {
						#print "not defined\n";
					}
				}
			}
			if ((defined $atom_assignment->{'label'}) and $atom_assignment->{'label'} =~ /(val)/i) {
				if ($name =~ /(cg)(\d)$/i) {
					my $ambig = $atom_assignment->{'ambig'};
					foreach my $key (%$residue_assignment) {
						#print "key: $key and looking for: HG$2\n";
					}
					if (exists $residue_assignment->{"HG" . $2}) {
						my $methyl_proton_assignment = $residue_assignment->{"HG" . $1};
						$residue_assignment->{"HG" . $2}->{'ambig'} = $ambig;
						$self->debug("setting ambig code at $id for HG$2 to $ambig");
					} else {
						#print "not defined\n";
					}
				}
			}
		}
	}
	foreach my $seqcode (sort{$a<=>$b}(keys (%chem_shifts))) {
		my $residue_assignment = $chem_shifts{$seqcode};
		foreach my $atom_assignment (sort{$a->{'id'}<=>$b->{'id'}}(values (%$residue_assignment))) {
			my $id = $atom_assignment->{'id'};
			my $label = $atom_assignment->{'label'};
			my $name = $atom_assignment->{'name'};
			my $type = $atom_assignment->{'type'};
			my $value = $atom_assignment->{'value'};
			my $error = $atom_assignment->{'error'};
			my $ambig =  $atom_assignment->{'ambig'};
			if (exists $self->{pseudoatom}->{uc($label . $name)}) {
				$name .= $self->{pseudoatom}->{uc($label . $name)};
			}
			push(@stk, "$id $seqcode $label $name $type $value $error $ambig");
		}
	}
	return \@stk;
}
sub parse_bmrb {
	my $self = shift;
	my $chem_shifts_ref = $self->{extracted_chemical_shifts};
	my %chem_shifts = %$chem_shifts_ref;
	$self->debug("parsing bmrb: " . $self->get_filepath());
	my $temp_line = "";
	open ( BMRBFILE, "<" . $self->get_filepath() ) or die "$!";
	binmode BMRBFILE;
	while (defined BMRBFILE and $temp_line = <BMRBFILE>) {
		chomp($temp_line);
		next if (($temp_line =~ /^\#/) or ($temp_line =~ /^\s*$/));
#		$self->debug("$temp_line");
		if ($temp_line =~ /save_([^\s]+)/i) {
			my @save = ($temp_line);
			my $save_category;
			while ((defined BMRBFILE) and ($temp_line = <BMRBFILE>) and not $temp_line =~ /save_\s*$/) {
				next if (($temp_line =~ /^\#/) or ($temp_line =~ /^\s*$/));
#				$self->debug("**$temp_line");
				$temp_line =~ s/^\s+//;
#				$temp_line =~ s/\s+$//;
				push(@save, $temp_line);
				if ($temp_line =~ /_saveframe_category\s+([^\s]+)/i or $temp_line =~ /sf_category\s+([^\s]+)/i) {
 					$save_category = $1;
					$self->debug("found save category " . $save_category);;
				}
			}
			if ($save_category =~ /monomeric_polymer/i) {
				$self->{version} = "2.x";
				$self->{monomeric_polymer_frame} = \@save;
			} elsif ($save_category =~ /entity/i) {
				$self->debug("entity saveframe found");
				$self->{version} = "3.x";
				$self->{assigned_chemical_shifts_frame} = \@save;
				$self->{monomeric_polymer_frame} = \@save;
			} elsif ($save_category =~ /assigned_chemical_shifts/i) {
				$self->{assigned_chemical_shifts_frame} = \@save;
			}
		} else {
			#test here for non bmrb format
			$temp_line =~ s/^\s+//;
			$temp_line =~ s/\s+$//;
			my @items = split(/\s+/, $temp_line);
			if (scalar(@items) == 8) {
				if ($temp_line =~ /(\d+)\s+(\d+)\s+([a-zA-Z]{3}\S*)\s+(\w+)\s+([a-zA-Z])\s+(\S+)\s+(\S+)\s+(\S+)/) {
					my $seqcode = $2;
					my $label = $4;
					$self->{residues}->{$seqcode} = $3;
					unless (defined $chem_shifts{$seqcode}) {
						$chem_shifts{$seqcode} = {};
						$self->debug("new residue $seqcode ");
					} else {
						$self->debug("old residue $seqcode ");
					}
					my $assigned_atoms_ref = $chem_shifts{$seqcode};
					$assigned_atoms_ref->{$label} = {id => "$1", label => "$3", name => uc("$label"), type => "$5", value => "$6", error => "$7", ambig => "$8"};
					$k++;
					$self->debug("adding: *$1* *$2* *$3* *$label* *$5* *$6* *$7* *$8*");
					$self->{version} = "autostructure";
				}
			} else {
				$self->debug("** skipping line: $temp_line");
			}
		}
	}
	close BMRBFILE;
	$self->{extracted_chemical_shifts} = \%chem_shifts;
	defined ($self->{version}) or die "## The file is not in a supported BMRB format.\n## Supported versions are 2.x and 3.x.\n## Please visit http://www.bmrb.wisc.edu/formats.html for more details.\n";
}
sub parse_save {
	my $self = shift;
	my $category = shift;
	my @save = shift;
	$self->debug($save[0] . " ,cat: " . $category);
	foreach my $line (@save) {
		while (not $temp_line =~ /loop_$/) {
			next if (($temp_line =~ /^\#/) or ($temp_line =~ /^\s*$/));
			push(@save, $temp_line);
			$temp_line = <BMRBFILE>;
		}
		$self->parse_save(@loop);
	}
}
sub get_assigned_chemical_shifts_frame {
	my $self = shift;
	return $self->{assigned_chemical_shifts_frame};
}
sub get_monomeric_polymer_frame {
	my $self = shift;
	return $self->{monomeric_polymer_frame};
}
sub get_filename {
	my $self = shift;
	return $self->{filename};
}
sub set_filename {
	my $self = shift;
	$self->{filename} = shift;
}
sub get_filepath {
	my $self = shift;
	return $self->{filepath};
}
sub set_filepath {
	my $self = shift;
	$self->{filepath} = shift;
}
sub debug {
	my $self = shift;
	if ($self->{debug}) {
		print "@_\n";
	}
}
return(1); 
 